// Copyright (C) 2022 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'package:example_router/core/constants/form_field.dart';
import 'package:example_router/core/utils/custom_password.dart';
import 'package:flutter/material.dart' hide FormField;
import 'package:wyatt_authentication_bloc/wyatt_authentication_bloc.dart';
import 'package:wyatt_form_bloc/wyatt_form_bloc.dart';

class _EmailInput extends StatelessWidget {
  @override
  Widget build(BuildContext context) {
    return InputBuilder<SignUpCubit<int>>(
      field: AuthFormField.email,
      builder: ((context, cubit, state, field, input) {
        return TextField(
          onChanged: (email) => cubit.emailChanged(email),
          keyboardType: TextInputType.emailAddress,
          decoration: InputDecoration(
            labelText: 'Email',
            helperText: '',
            errorText: input.validator.invalid ? 'Invalid email' : null,
          ),
        );
      }),
    );
  }
}

class _PasswordInput extends StatelessWidget {
  @override
  Widget build(BuildContext context) {
    return InputBuilder<SignUpCubit<int>>(
      field: AuthFormField.password,
      builder: ((context, cubit, state, field, input) {
        return TextField(
          onChanged: (pwd) {
            cubit.passwordCustomChanged<CustomPassword>(
                CustomPassword.dirty(pwd));
            cubit.dataChanged(
                AppFormField.confirmedPassword,
                ConfirmedPassword.dirty(
                    password: pwd,
                    value: state.form
                        .valueOf<String?>(AppFormField.confirmedPassword)));
          },
          obscureText: true,
          decoration: InputDecoration(
            labelText: 'Password',
            helperText: '',
            errorText: input.validator.invalid ? 'Invalid password' : null,
          ),
        );
      }),
    );
  }
}

class _ConfirmPasswordInput extends StatelessWidget {
  @override
  Widget build(BuildContext context) {
    return InputBuilder<SignUpCubit<int>>(
      field: AppFormField.confirmedPassword,
      builder: ((context, cubit, state, field, input) {
        return TextField(
          onChanged: (pwd) {
            cubit.dataChanged(
              field,
              ConfirmedPassword.dirty(
                  password:
                      state.form.valueOf<String?>(AuthFormField.password) ?? '',
                  value: pwd),
            );
          },
          obscureText: true,
          decoration: InputDecoration(
            labelText: 'Confirm password',
            helperText: '',
            errorText:
                input.validator.invalid ? 'Passwords do not match' : null,
          ),
        );
      }),
    );
  }
}

class _SignUpButton extends StatelessWidget {
  @override
  Widget build(BuildContext context) {
    return SubmitBuilder<SignUpCubit<int>>(
      builder: ((context, cubit, status) {
        return status.isSubmissionInProgress
            ? const CircularProgressIndicator()
            : ElevatedButton(
                onPressed: status.isValidated ? () => cubit.submit() : null,
                child: const Text('Sign up'),
              );
      }),
    );
  }
}

class SignUpForm extends StatelessWidget {
  const SignUpForm({Key? key}) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return SignUpListener<int>(
        onError: (context, status, errorMessage) =>
            ScaffoldMessenger.of(context)
              ..hideCurrentSnackBar()
              ..showSnackBar(
                SnackBar(content: Text(errorMessage ?? 'Sign Up Failure')),
              ),
        child: SingleChildScrollView(
          child: Column(
            children: [
              _EmailInput(),
              const SizedBox(height: 8),
              _PasswordInput(),
              const SizedBox(height: 8),
              _ConfirmPasswordInput(),
              const SizedBox(height: 16),
              _SignUpButton(),
            ],
          ),
        ));
  }
}
